"""OpenCV windows that will display the arrays."""

import warnings
from threading import Thread
from typing import List, Union, Callable, Any, Dict, Iterable, Optional

import cv2
import numpy as np
from localpubsub import NoData

from displayarray.callbacks import global_cv_display_callback
from displayarray._uid import uid_for_source
from displayarray.frame import subscriber_dictionary
from displayarray.frame.frame_updater import FrameCallable
from displayarray.frame.frame_updater import FrameUpdater
from displayarray.frame.subscriber_dictionary import CV_CAMS_DICT
from displayarray.input import MouseEvent
from displayarray.window import window_commands
from displayarray._util import WeakMethod
from displayarray.effects.select_channels import SelectChannels

try:
    import sys

    if sys.platform == "linux":
        from PyV4L2Cam.get_camera import get_bus_info_from_camera  # type: ignore
    else:
        get_bus_info_from_camera = None
except:
    get_bus_info_from_camera = None


class SubscriberWindows(object):
    """Windows that subscribe to updates to cameras, videos, and arrays."""

    FRAME_DICT: Dict[str, np.ndarray] = {}
    ESC_KEY_CODES = [27]  # ESC key on most keyboards

    def __init__(
        self,
        window_names: Iterable[str] = ("displayarray",),
        video_sources: Iterable[Union[str, int]] = (0,),
        callbacks: Optional[List[Callable[[np.ndarray], Any]]] = None,
        silent: bool = False,
    ):
        """Create the array displaying window."""
        self.source_names: List[Union[str, int]] = []
        self.close_threads: Optional[List[Thread]] = []
        self.frames: Dict[Union[str, int], np.ndarray] = {}
        self.input_vid_global_names: List[str] = []
        self.window_names: List[str] = []
        self.input_cams: List[str] = []
        self.exited = False
        self.silent = silent

        if callbacks is None:
            callbacks = []
        for name in video_sources:
            self.add_source(name)
        self.callbacks = callbacks
        if not self.silent:
            for name in window_names:
                self.add_window(name)

        self.update()

    def __bool__(self):
        self.update()
        return not self.exited

    def __iter__(self):
        while not self.exited:
            self.update()
            yield self.frames

    def block(self):
        """Update the window continuously while blocking the outer program."""
        self.loop()
        for ct in self.close_threads:
            ct.join()

    def add_source(self, name):
        """Add another source for this class to display."""
        uid = uid_for_source(name)
        self.source_names.append(uid)
        self.input_vid_global_names.append(uid)
        self.input_cams.append(name)
        return self

    def add_window(self, name):
        """Add another window for this class to display sources with. The name will be the title."""
        self.window_names.append(name)
        cv2.namedWindow(name + " (press ESC to quit)")
        m = WeakMethod(self.handle_mouse)
        cv2.setMouseCallback(name + " (press ESC to quit)", m)
        return self

    def add_callback(self, callback):
        """Add a callback for this class to apply to videos."""
        self.callbacks.append(callback)
        return self

    def __stop_all_cams(self):
        for c in self.source_names:
            subscriber_dictionary.stop_cam(c)

    def handle_keys(self, key_input: int):
        """Capture key input for the escape function and passing to key control subscriber threads."""
        if key_input in self.ESC_KEY_CODES:
            for name in self.window_names:
                cv2.destroyWindow(name + " (press ESC to quit)")
            self.exited = True
            window_commands.quit()
            self.__stop_all_cams()
            return "quit"
        elif key_input not in [-1, 0]:
            try:
                window_commands.key_pub.publish(chr(key_input))
            except ValueError:
                warnings.warn(
                    RuntimeWarning(
                        f"Unknown key code: [{key_input}]. Please report to the displayarray issue page."
                    )
                )

    def handle_mouse(self, event, x, y, flags, param):
        """Capture mouse input for mouse control subscriber threads."""
        mousey = MouseEvent(event, x, y, flags, param)
        window_commands.mouse_pub.publish(mousey)

    def display_frames(self, frames, win_num=0, prepend_name=""):
        """Display a list of frames on multiple windows."""
        if isinstance(frames, Exception):
            raise frames
        if isinstance(frames, dict):
            for f_name, f in frames.items():
                for i in range(len(f)):
                    # detect nested:
                    if (
                        isinstance(f[i], (list, tuple))
                        or f[i].dtype.num == 17
                        or (
                            len(f[i].shape) != 2
                            and (len(f[i].shape) != 3 or f[i].shape[-1] != 3)
                        )
                    ):
                        win_num = self.display_frames(
                            f[i], win_num, prepend_name=f"{f_name} - "
                        )
                    else:
                        if len(self.window_names) <= win_num:
                            self.add_window(f"{prepend_name}{win_num}")
                        cv2.imshow(
                            self.window_names[win_num] + " (press ESC to quit)", f[i]
                        )
                        win_num += 1
        else:
            for f in range(len(frames)):
                # detect nested:
                if (
                    isinstance(frames[f], (list, tuple))
                    or frames[f].dtype.num == 17
                    or (
                        len(frames[f].shape) != 2
                        and (len(frames[f].shape) != 3 or frames[f].shape[-1] != 3)
                    )
                ):
                    win_num = self.display_frames(frames[f], win_num, prepend_name)
                else:
                    if len(self.window_names) <= win_num:
                        self.add_window(f"{prepend_name} {win_num}")
                    cv2.imshow(
                        self.window_names[win_num] + " (press ESC to quit)", frames[f]
                    )
                    win_num += 1
        return win_num

    def __check_too_many_channels(self):
        for f_name, f in self.frames.items():
            for i in range(len(f)):
                if isinstance(f[i], Exception):
                    raise f[i]
                if (
                    isinstance(f[i], np.ndarray)
                    and f[i].shape[-1] not in [1, 3]
                    and len(f[i].shape) != 2
                ):
                    print(
                        f"Too many channels in output. (Got {f[i].shape[-1]} instead of 1 or 3.) "
                        f"Frame selection callback added."
                    )
                    print(
                        "Ctrl+scroll to change first channel.\n"
                        "Shift+scroll to change second channel.\n"
                        "Alt+scroll to change third channel."
                    )
                    sel = SelectChannels()
                    sel.enable_mouse_control()
                    sel.mouse_print_channels = True
                    self.callbacks.append(sel)
                    for fr in range(len(f)):
                        f[fr] = self.callbacks[-1](f[fr])
                    break

    def update_frames(self):
        """Update the windows with the newest data for all frames."""
        self.frames = {}

        for i in range(len(self.input_vid_global_names)):
            if self.input_vid_global_names[i] in self.FRAME_DICT and not isinstance(
                self.FRAME_DICT[self.input_vid_global_names[i]], NoData
            ):
                if self.input_vid_global_names[i] not in self.frames.keys():
                    self.frames[self.input_vid_global_names[i]] = []
                self.frames[self.input_vid_global_names[i]].append(
                    self.FRAME_DICT[self.input_vid_global_names[i]]
                )
                if len(self.callbacks) > 0:
                    for c in self.callbacks:
                        frame = c(self.frames[self.input_vid_global_names[i]][-1])
                        if frame is not None:
                            self.frames[self.input_vid_global_names[i]][-1] = frame
                if not self.silent:
                    self.__check_too_many_channels()
        if not self.silent:
            self.display_frames(self.frames)

    def update(self, arr: np.ndarray = None, id: str = None):
        """Update window frames once. Optionally add a new input and input id."""
        if arr is not None and id is not None:
            global_cv_display_callback(arr, id)
            if id not in self.input_cams:
                self.add_source(id)
                if not self.silent:
                    self.add_window(id)
        sub_cmd = window_commands.win_cmd_sub()
        self.update_frames()
        msg_cmd = sub_cmd.get()
        key = self.handle_keys(cv2.waitKey(1))
        return msg_cmd, key

    def wait_for_init(self):
        """Update window frames in a loop until they're actually updated. Useful for waiting for cameras to init."""
        msg_cmd = ""
        key = ""
        while msg_cmd != "quit" and key != "quit" and len(self.frames) == 0:
            msg_cmd, key = self.update()
        return self

    def end(self):
        """Close all threads. Should be used with non-blocking mode."""
        window_commands.quit(force_all_read=False)
        self.__stop_all_cams()
        for t in self.close_threads:
            t.join()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.end()

    def __del__(self):
        self.end()

    def __delete__(self, instance):
        self.end()

    def loop(self):
        """Continually update window frame. OpenCV only allows this in the main thread."""
        sub_cmd = window_commands.win_cmd_sub()
        msg_cmd = ""
        key = ""
        while msg_cmd != "quit" and key != "quit":
            msg_cmd, key = self.update()
        sub_cmd.release()
        window_commands.quit(force_all_read=False)
        self.__stop_all_cams()

    @property
    def cams(self):
        """Get the camera instances. Can be used for OpenCV or V4L2 functions, depending on backend."""
        return [CV_CAMS_DICT[v].cam_instance for v in self.input_vid_global_names]

    @property
    def busses(self):
        """Get the busses the cameras are plugged into. Can be used as UIDs. Requires V4L2 backend."""
        if get_bus_info_from_camera is not None:
            return [get_bus_info_from_camera(c) for c in self.cams]
        else:
            raise RuntimeError("Getting bus info not supported on this system")


def _get_video_callback_dict_threads(
    *vids,
    callbacks: Optional[Dict[Any, Union[FrameCallable, List[FrameCallable]]]] = None,
    fps=float("inf"),
    size=(-1, -1),
    force_backend="",
):
    assert callbacks is not None
    vid_threads = []
    for v in vids:
        v_name = uid_for_source(v)
        v_callbacks: List[Callable[[np.ndarray], Any]] = []
        if v_name in callbacks:
            if isinstance(callbacks[v_name], List):
                v_callbacks.extend(callbacks[v_name])  # type: ignore
            elif callable(callbacks[v_name]):
                v_callbacks.append(callbacks[v_name])  # type: ignore
        if v in callbacks:
            if isinstance(callbacks[v], List):
                v_callbacks.extend(callbacks[v])  # type: ignore
            elif callable(callbacks[v]):
                v_callbacks.append(callbacks[v])  # type: ignore
        vid_threads.append(
            FrameUpdater(
                v,
                callbacks=v_callbacks,
                fps_limit=fps,
                request_size=size,
                force_backend=force_backend,
            )
        )
    return vid_threads


def _get_video_threads(
    *vids,
    callbacks: Optional[
        Union[
            Dict[Any, Union[FrameCallable, List[FrameCallable]]],
            List[FrameCallable],
            FrameCallable,
        ]
    ] = None,
    fps=float("inf"),
    size=(-1, -1),
    force_backend="",
):
    vid_threads: List[Thread] = []
    if isinstance(callbacks, Dict):
        vid_threads = _get_video_callback_dict_threads(
            *vids, callbacks=callbacks, fps=fps, size=size
        )
    elif isinstance(callbacks, List):
        for v in vids:
            vid_threads.append(
                FrameUpdater(
                    v,
                    callbacks=callbacks,
                    fps_limit=fps,
                    request_size=size,
                    force_backend=force_backend,
                )
            )
    elif callable(callbacks):
        for v in vids:
            vid_threads.append(
                FrameUpdater(
                    v,
                    callbacks=[callbacks],
                    fps_limit=fps,
                    request_size=size,
                    force_backend=force_backend,
                )
            )
    else:
        for v in vids:
            if v is not None:
                vid_threads.append(
                    FrameUpdater(
                        v, fps_limit=fps, request_size=size, force_backend=force_backend
                    )
                )
    return vid_threads


def display(
    *vids,
    callbacks: Optional[
        Union[
            Dict[Any, Union[FrameCallable, List[FrameCallable]]],
            List[FrameCallable],
            FrameCallable,
        ]
    ] = None,
    window_names=None,
    blocking=False,
    fps_limit=float("inf"),
    size=(-1, -1),
    silent=False,
    force_backend="",
):
    """
    Display all the arrays, cameras, and videos passed in.

    callbacks can be a dictionary linking functions to videos, or a list of function or functions operating on the video
     data before displaying.
    Window names end up becoming the title of the windows
    """
    vid_threads = _get_video_threads(
        *vids,
        callbacks=callbacks,
        fps=fps_limit,
        size=size,
        force_backend=force_backend,
    )
    for v in vid_threads:
        v.start()
    if window_names is None:
        window_names = ["window {}".format(i) for i in range(len(vids))]
    if blocking:
        SubscriberWindows(
            window_names=window_names, video_sources=vids, silent=silent
        ).loop()
        for vt in vid_threads:
            vt.join()
    else:
        s = SubscriberWindows(
            window_names=window_names, video_sources=vids, silent=silent
        )
        s.close_threads = vid_threads
        return s


def breakpoint_display(*args, **kwargs):
    """Display all the arrays, cameras, and videos passed in. Stops code execution until the window is closed."""
    return display(*args, **kwargs, blocking=True)


def read_updates(*args, **kwargs):
    """Read back all frame updates and yield a list of frames. List is empty if no frames were read."""
    return display(*args, **kwargs, silent=True)



"""
Display any array, webcam, or video file.

display is a function that displays these in their own windows.
"""

__version__ = "1.1.1"

from .window.subscriber_windows import display, breakpoint_display, read_updates
from .frame.frame_publishing import publish_updates_zero_mq, publish_updates_ros
from . import effects



"""Get and handle updated frames."""

import threading
import asyncio
from typing import Union, Tuple, Any, Callable, List, Optional, Dict

import numpy as np

from displayarray.callbacks import global_cv_display_callback
from displayarray._uid import uid_for_source
from displayarray.frame import subscriber_dictionary
from displayarray.frame.frame_publishing import pub_cam_thread
from displayarray.window import window_commands
from displayarray.effects.select_channels import SelectChannels
from localpubsub import NoData

FrameCallable = Callable[[np.ndarray], Optional[np.ndarray]]


class FrameUpdater(threading.Thread):
    """Thread for updating frames from a video source."""

    def __init__(
        self,
        video_source: Union[int, str, np.ndarray] = 0,
        callbacks: Optional[Union[List[FrameCallable], FrameCallable]] = None,
        request_size: Tuple[int, int] = (-1, -1),
        high_speed: bool = True,
        fps_limit: float = float("inf"),
        force_backend="",
    ):
        """Create the frame updater thread."""
        super(FrameUpdater, self).__init__(target=self.loop, args=())
        self.cam_id = uid_for_source(video_source)
        self.video_source = video_source
        if callbacks is None:
            callbacks = []
        if callable(callbacks):
            self.callbacks = [callbacks]
        else:
            self.callbacks = callbacks
        self.request_size = request_size
        self.high_speed = high_speed
        self.fps_limit = fps_limit
        self.exception_raised = None
        self.force_backend = force_backend

    def __wait_for_cam_id(self):
        while str(self.cam_id) not in subscriber_dictionary.CV_CAMS_DICT:
            continue

    def __apply_callbacks_to_frame(self, frame):
        if frame is not None and not isinstance(frame, NoData):
            try:
                for c in self.callbacks:
                    frame_c = c(frame)
                    if frame_c is not None:
                        frame = frame_c
                if (
                    isinstance(frame, np.ndarray)
                    and frame.shape[-1] not in [1, 3]
                    and len(frame.shape) != 2
                ):
                    print(
                        f"Too many channels in output. (Got {frame.shape[-1]} instead of 1 or 3.) "
                        f"Frame selection callback added."
                    )
                    print(
                        "Ctrl+scroll to change first channel.\n"
                        "Shift+scroll to change second channel.\n"
                        "Alt+scroll to change third channel."
                    )
                    sel = SelectChannels()
                    sel.enable_mouse_control()
                    sel.mouse_print_channels = True
                    self.callbacks.append(sel)
                    frame = self.callbacks[-1](frame)
            except Exception as e:
                self.exception_raised = e
                frame = self.exception_raised
                subscriber_dictionary.stop_cam(self.cam_id)
                window_commands.quit()
                raise e
            global_cv_display_callback(frame, self.cam_id)

    def loop(self):
        """Continually get frames from the video publisher, run callbacks on them, and listen to commands."""
        t = pub_cam_thread(
            self.video_source,
            self.request_size,
            self.high_speed,
            self.fps_limit,
            self.force_backend,
        )
        self.__wait_for_cam_id()

        sub_cam = subscriber_dictionary.cam_frame_sub(str(self.cam_id))
        sub_owner = subscriber_dictionary.handler_cmd_sub(str(self.cam_id))
        msg_owner = sub_owner.return_on_no_data = ""
        try:
            while msg_owner != "quit":
                frame = sub_cam.get(blocking=True, timeout=1.0)  # type: np.ndarray
                self.__apply_callbacks_to_frame(frame)
                msg_owner = sub_owner.get()
        except Exception as e:
            raise e
        finally:
            sub_owner.release()
            sub_cam.release()
            subscriber_dictionary.stop_cam(self.cam_id)
            t.join()

    def display(self, callbacks: List[Callable[[np.ndarray], Any]] = None):
        """
        Start default display operation.

        For multiple video sources, please use something outside of this class.

        :param callbacks: List of callbacks to be run on frames before displaying to the screen.
        """
        from displayarray.window import SubscriberWindows

        if callbacks is None:
            callbacks = []
        self.start()
        SubscriberWindows(video_sources=[self.cam_id], callbacks=callbacks).loop()
        self.join()
        if self.exception_raised is not None:
            raise self.exception_raised


'''async def read_updates(
    *vids,
    callbacks: Optional[
        Union[
            Dict[Any, Union[FrameCallable, List[FrameCallable]]],
            List[FrameCallable],
            FrameCallable,
        ]
    ] = None,
    fps_limit=float("inf"),
    size=(-1, -1),
    end_callback: Callable[[], bool] = lambda: False,
    blocking=True,
):
    """
    Read back all updates from the requested videos.

    Examp#le usage:

    .. co#de-block:: python

      >>#> from examples.videos import test_video
      >>#> f = 0
      >>#> for f, r in enumerate(read_updates(test_video, end_callback=lambda :f==2)):
      ..#.   print(f"Frame:{f}. Array:{r}")

    """
    from displayarray.window import SubscriberWindows
    from displayarray.window.subscriber_windows import _get_video_threads

    vid_names = [uid_for_source(name) for name in vids]
    vid_threads = _get_video_threads(
        *vids, callbacks=callbacks, fps=fps_limit, size=size
    )
    for v in vid_threads:
        v.start()

    while not end_callback():
        vid_update_dict = {}
        dict_was_updated = False
        for i in range(len(vid_names)):
            if vid_names[i] in SubscriberWindows.FRAME_DICT and not isinstance(
                SubscriberWindows.FRAME_DICT[vid_names[i]], NoData
            ):
                vid_update_dict[vid_names[i]] = SubscriberWindows.FRAME_DICT[
                    vid_names[i]
                ]
                if (
                    isinstance(vid_update_dict[vid_names[i]], np.ndarray)
                    and len(vid_update_dict[vid_names[i]].shape) <= 3
                ):
                    vid_update_dict[vid_names[i]] = [vid_update_dict[vid_names[i]]]
                dict_was_updated = True
        if dict_was_updated or not blocking:
            yield vid_update_dict
        await asyncio.sleep(0)
    for v in vid_names:
        subscriber_dictionary.stop_cam(v)
    for v in vid_threads:
        v.join()'''


async def read_updates_zero_mq(
    *topic_names,
    address: str = "tcp://127.0.0.1:5600",
    flags: int = 0,
    copy: bool = True,
    track: bool = False,
    blocking: bool = False,
    end_callback: Callable[[Any], bool] = lambda x: False,
):
    """Read updated frames from ZeroMQ."""
    import zmq

    ctx = zmq.Context()
    s = ctx.socket(zmq.SUB)
    s.connect(address)
    if not blocking:
        flags |= zmq.NOBLOCK

    for topic in topic_names:
        s.setsockopt(zmq.SUBSCRIBE, topic)
    cb_val = False
    while not cb_val:
        try:
            md = s.recv_json(flags=flags)
            msg = s.recv(flags=flags, copy=copy, track=track)
            buf = memoryview(msg)
            arr = np.frombuffer(buf, dtype=md["dtype"])
            arr.reshape(md["shape"])
            name = md["name"]
            cb_val = end_callback(md)
            yield name, arr
        except zmq.ZMQError as e:
            if isinstance(e, zmq.Again):
                pass  # no messages to receive
            else:
                raise e
        finally:
            await asyncio.sleep(0)


async def read_updates_ros(
    *topic_names,
    dtypes=None,
    listener_node_name=None,
    poll_rate_hz=None,
    end_callback: Callable[[Any], bool] = lambda x: False,
):
    """Read updated frames from ROS."""
    import rospy
    from rospy.numpy_msg import numpy_msg
    from rospy.client import _WFM
    import std_msgs.msg
    import random
    import string

    if dtypes is None:
        raise ValueError(
            "ROS cannot automatically determine the types of incoming numpy arrays. Please specify.\n"
            "Options are: \n"
            "\tfloat32, float64, bool, char, int16, "
            "\tint32, int64, str, uint16, uint32, uint64, uint8"
        )

    if listener_node_name is None:
        # https://stackoverflow.com/a/2257449
        listener_node_name = "".join(
            random.choices(string.ascii_uppercase + string.digits, k=8)
        )

    rospy.init_node(listener_node_name)

    msg_types = [
        {
            np.float32: std_msgs.msg.Float32(),
            np.float64: std_msgs.msg.Float64(),
            np.bool: std_msgs.msg.Bool(),
            np.char: std_msgs.msg.Char(),
            np.int16: std_msgs.msg.Int16(),
            np.int32: std_msgs.msg.Int32(),
            np.int64: std_msgs.msg.Int64(),
            np.str: std_msgs.msg.String(),
            np.uint16: std_msgs.msg.UInt16(),
            np.uint32: std_msgs.msg.UInt32(),
            np.uint64: std_msgs.msg.UInt64(),
            np.uint8: std_msgs.msg.UInt8(),
            "float32": std_msgs.msg.Float32(),
            "float64": std_msgs.msg.Float64(),
            "bool": std_msgs.msg.Bool(),
            "char": std_msgs.msg.Char(),
            "int16": std_msgs.msg.Int16(),
            "int32": std_msgs.msg.Int32(),
            "int64": std_msgs.msg.Int64(),
            "str": std_msgs.msg.String(),
            "uint16": std_msgs.msg.UInt16(),
            "uint32": std_msgs.msg.UInt32(),
            "uint64": std_msgs.msg.UInt64(),
            "uint8": std_msgs.msg.UInt8(),
        }.get(
            dtype, dtype
        )  # allow users to use their own custom messages in numpy arrays
        for dtype in dtypes
    ]
    s = None
    cb_val = False
    try:
        wfms = {t: _WFM() for t in topic_names}
        s = {
            t: rospy.Subscriber(t, numpy_msg(msg_types[i]), wfms[t].cb)
            for i, t in enumerate(topic_names)
        }
        while not cb_val:
            while not rospy.core.is_shutdown():
                if poll_rate_hz:
                    await asyncio.sleep(1.0 / poll_rate_hz)
                else:
                    await asyncio.sleep(0)
                for t, w in wfms.items():
                    if w.msg is not None:
                        yield t, w.msg
                        cb_val = end_callback(w.msg)
                        w.msg = None
    except KeyboardInterrupt:
        pass
    finally:
        if s is not None:
            for _, sub in s.items():
                sub.unregister()
    if rospy.core.is_shutdown():
        raise rospy.exceptions.ROSInterruptException("rospy shutdown")



"""Decorators for creating input loops that OpenCV handles."""

from displayarray.window import window_commands
import threading
import time

from typing import Callable


class MouseEvent(object):
    """Holds all the OpenCV mouse event information."""

    def __init__(self, event, x, y, flags, param):
        """Create an OpenCV mouse event."""
        self.event = event
        self.x = x
        self.y = y
        self.flags = flags
        self.param = param

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        return "event:{}\nx,y:{},{}\nflags:{}\nparam:{}\n".format(
            self.event, self.x, self.y, self.flags, self.param
        )


class _mouse_thread(object):  # NOSONAR
    """Run a function on mouse information that is received by the window."""

    def __init__(self, f):
        self.f = f
        self.sub_mouse = window_commands.mouse_pub.make_sub()

    def __call__(self, *args, **kwargs):
        """Call the function this was set up with."""
        self.f(self.sub_mouse, *args, **kwargs)


class _mouse_loop_thread(object):  # NOSONAR
    """Run a function on mouse information that is received by the window, in the main loop."""

    def __init__(self, f, run_when_no_events=False, fps=60):
        self.f = f
        self.sub_mouse = window_commands.mouse_pub.make_sub()
        self.sub_cmd = window_commands.win_cmd_pub.make_sub()
        self.sub_cmd.return_on_no_data = ""
        self.run_when_no_events = run_when_no_events
        self.fps = fps

    def __call__(self, *args, **kwargs):
        """Run the function this was set up with in a loop."""
        msg_cmd = ""
        while msg_cmd != "quit":
            mouse_xyzclick = self.sub_mouse.get(blocking=True)  # type: MouseEvent
            if mouse_xyzclick is not self.sub_mouse.return_on_no_data:
                self.f(mouse_xyzclick, *args, **kwargs)
            elif self.run_when_no_events:
                self.f(None, *args, **kwargs)
            msg_cmd = self.sub_cmd.get()
            time.sleep(1.0 / self.fps)
        window_commands.quit(force_all_read=False)


class mouse_loop(object):  # NOSONAR
    """
    Run a function on mouse information that is received by the window, continuously in a new thread.

    .. code-block:: python

      >>> @mouse_loop
      ... def fun(mouse_event):
      ...   print("x:{}, y:{}".format(mouse_event.x, mouse_event.y))
    """

    def __init__(self, f):
        """Start a new mouse thread for the decorated function."""
        self.t = threading.Thread(
            target=_mouse_loop_thread(f, run_when_no_events=False)
        )
        self.t.start()

    def __call__(self, *args, **kwargs):
        """Return the thread that was started with the function passed in."""
        return self.t


class _key_thread(object):  # NOSONAR
    """Run a function on mouse information that is received by the window."""

    def __init__(self, f):
        self.f = f
        self.sub_key = window_commands.key_pub.make_sub()

    def __call__(self, *args, **kwargs):
        """Call the function this was set up with."""
        self.f(self.sub_key, *args, **kwargs)


class _key_loop_thread(object):  # NOSONAR
    """Run a function on mouse information that is received by the window, in the main loop."""

    def __init__(self, f, run_when_no_events=False, fps=60):
        self.f = f
        self.sub_key = window_commands.key_pub.make_sub()
        self.sub_cmd = window_commands.win_cmd_pub.make_sub()
        self.sub_cmd.return_on_no_data = ""
        self.run_when_no_events = run_when_no_events
        self.fps = fps

    def __call__(self, *args, **kwargs):
        """Run the function this was set up with in a loop."""
        msg_cmd = ""
        while msg_cmd != "quit":
            key_chr = self.sub_key.get()  # type: chr
            if key_chr is not self.sub_key.return_on_no_data:
                self.f(key_chr, *args, **kwargs)
            elif self.run_when_no_events:
                self.f(None, *args, **kwargs)
            msg_cmd = self.sub_cmd.get()
            time.sleep(1.0 / self.fps)
        window_commands.quit(force_all_read=False)


class key_loop(object):  # NOSONAR
    """
    Run a function on mouse information that is received by the window, continuously in a new thread.

    .. code-block:: python

      >>> @key_loop
      ... def fun(key):
      ...   print("key pressed:{}".format(key))
    """

    def __init__(self, f: Callable[[str], None]):
        """Start a new key thread for the decorated function."""
        self.t = threading.Thread(target=_key_loop_thread(f, run_when_no_events=False))
        self.t.start()

    def __call__(self, *args, **kwargs):
        """Return the thread that was started with the function passed in."""
        return self.t



"""Publisher-subscriber commands to and from the camera."""

from localpubsub import VariablePub, VariableSub

from typing import Union, Dict


class CamHandler(object):
    """A camera handler instance that will send commands to and receive data from a camera."""

    def __init__(self, name, sub):
        """Create the cam handler."""
        self.name = name
        self.cmd = None
        self.sub: VariableSub = sub
        self.pub = VariablePub()
        self.cmd_pub = VariablePub()


class Cam(object):
    """A camera publisher instance that will send frames, status, and commands out."""

    def __init__(self, name, cam_instance=None):
        """Create the cam."""
        self.name = name
        self.cmd = None
        self.frame_pub = VariablePub()
        self.cmd_pub = VariablePub()
        self.status_pub = VariablePub()
        self.cam_instance = cam_instance


CV_CAM_HANDLERS_DICT: Dict[str, CamHandler] = {}
CV_CAMS_DICT: Dict[str, Cam] = {}


def register_cam(cam_id, cam_instance=None):
    """Register camera "cam_id" to a global list so it can be picked up."""
    cam = Cam(str(cam_id), cam_instance)
    CV_CAMS_DICT[str(cam_id)] = cam
    CV_CAM_HANDLERS_DICT[str(cam_id)] = CamHandler(
        str(cam_id), cam.frame_pub.make_sub()
    )


def stop_cam(cam_id: Union[int, str]):
    """Tell camera "cam_id" to end it's main loop."""
    if str(cam_id) in CV_CAMS_DICT:
        CV_CAMS_DICT[str(cam_id)].cmd_pub.publish("quit", blocking=True)
    if str(cam_id) in CV_CAM_HANDLERS_DICT:
        CV_CAM_HANDLERS_DICT[str(cam_id)].cmd_pub.publish("quit", blocking=True)


def cam_cmd_sub(cam_id, blocking=True):
    """Get a command subscriber for registered camera "cam_id"."""
    if blocking:
        while cam_id not in CV_CAMS_DICT:
            continue
    return CV_CAMS_DICT[str(cam_id)].cmd_pub.make_sub()


def cam_frame_sub(cam_id, blocking=True):
    """Get a frame subscriber for registered camera "cam_id"."""
    if blocking:
        while cam_id not in CV_CAMS_DICT:
            continue
    return CV_CAMS_DICT[str(cam_id)].frame_pub.make_sub()


def cam_status_sub(cam_id, blocking=True):
    """Get a status subscriber for registered camera "cam_id"."""
    if blocking:
        while cam_id not in CV_CAMS_DICT:
            continue
    return CV_CAMS_DICT[str(cam_id)].status_pub.make_sub()


def handler_cmd_sub(cam_id, blocking=True):
    """Get a command subscriber for registered camera "cam_id" handler."""
    if blocking:
        while cam_id not in CV_CAM_HANDLERS_DICT:
            continue
    return CV_CAM_HANDLERS_DICT[str(cam_id)].cmd_pub.make_sub()



"""Overlay functions."""


def overlay_transparent(background, overlay, x=None, y=None):
    """
    Overlay a transparent image on top of a background image.

    :param background: background rgb image to overlay on top of
    :param overlay: rgba image to overlay on top of the background
    :param x: leftmost part to overlay at on the background
    :param y: topmost part to overlay at on the background
    """
    # https://stackoverflow.com/a/54058766/782170
    assert overlay.shape[2] == 4, "Overlay must be BGRA"

    background_width = background.shape[1]
    background_height = background.shape[0]

    if (x is not None and x >= background_width) or (
        y is not None and y >= background_height
    ):
        return background

    h, w = overlay.shape[0], overlay.shape[1]

    if x is None:
        x = int(background_width / 2 - w / 2)
    if y is None:
        y = int(background_height / 2 - h / 2)

    if x < 0:
        w += x
        overlay = overlay[:, -x:]

    if y < 0:
        w += y
        overlay = overlay[:, -y:]

    if x + w > background_width:
        w = background_width - x
        overlay = overlay[:, :w]

    if y + h > background_height:
        h = background_height - y
        overlay = overlay[:h]

    overlay_image = overlay[..., :3]
    mask = overlay[..., 3:] / 255.0

    background[y : y + h, x : x + w] = (1.0 - mask) * background[
        y : y + h, x : x + w
    ] + mask * overlay_image

    return background



"""Generate unique IDs for videos."""

from collections.abc import Hashable


def uid_for_source(video_source):
    """Get a uid for any source so it can be passed through the publisher-subscriber system."""
    if len(str(video_source)) <= 1000:
        uid = str(video_source)
    elif isinstance(video_source, Hashable):
        try:
            uid = str(hash(video_source))
        except TypeError:
            raise NotImplementedError(
                "Displaying immutables filled with mutables is not allowed yet. "
                "No tuples of arrays."
            )
    else:
        uid = str(hash(str(video_source)))
    return uid



"""Create lens effects. Currently only 2D+color arrays are supported."""

import numpy as np
from ..input import mouse_loop
import cv2

try:
    import torch
except ImportError:
    torch = None  # type: ignore


class ControllableLens(object):
    """A lens callback that can be controlled by the program or the user."""

    def __init__(self, use_bleed=False, zoom=1, center=None):
        """Create the lens callback."""
        self._center = center
        self._zoom = zoom
        self.use_bleed = use_bleed
        self.bleed = None
        self.mouse_control = None

    def _check_setup_bleed(self, arr):
        if not isinstance(self.bleed, np.ndarray) and self.use_bleed:
            self.bleed = np.zeros_like(arr)

    def run_bleed(self, arr, x, y):
        """Spread color outwards, like food coloring in water."""
        arr[y, ...] = (arr[(y + 1) % len(y), ...] + arr[(y - 1) % len(y), ...]) / 2
        arr[:, x, ...] = (
            arr[:, (x + 1) % len(x), ...] + arr[:, (x - 1) % len(x), ...]
        ) / 2


class Barrel(ControllableLens):
    """A barrel lens distortion callback."""

    def __init__(
        self, use_bleed=False, barrel_power=1, pincushion_power=1, zoom=1, center=None
    ):
        """Create the barrel lens distortion callback."""
        super().__init__(use_bleed, zoom, center)
        self._center = center
        self._zoom = zoom
        self.use_bleed = use_bleed
        self.bleed = None
        self._barrel_power = barrel_power
        self.mouse_control = None
        self.input_size = None

    @property
    def center(self):
        """Guarded get center. Limits to within input."""
        if self.input_size is not None:
            self._center[:] = [
                min(max(0, s), self.input_size[i]) for i, s in enumerate(self._center)
            ]
        return self._center

    @center.setter
    def center(self, setpoint):
        """Guarded set center. Limits to within input."""
        if self.input_size is not None:
            setpoint = [
                min(max(0, s), self.input_size[i]) for i, s in enumerate(setpoint)
            ]
        self._center[:] = setpoint

    @property
    def zoom(self):
        """Guarded zoom. Avoids divide by zero conditions."""
        if self._zoom == 0:
            return 1e-15
        else:
            return self._zoom

    @property
    def barrel_power(self):
        """Guarded barrel power. Avoids divide by zero conditions."""
        if self._barrel_power == 0:
            return 1e-15
        else:
            return self._barrel_power

    @barrel_power.setter
    def barrel_power(self, setpoint):
        """Set the barrel power."""
        self._barrel_power = setpoint

    @zoom.setter  # type: ignore
    def zoom(self, setpoint):
        """Set the zoom power."""
        self._zoom = setpoint

    def enable_mouse_control(self, crop_size=None):
        """
        Enable the default mouse controls.

        Move the mouse to center the image
        scroll to increase/decrease barrel
        ctrl+scroll to increase/decrease zoom
        """

        @mouse_loop
        def m_loop(me):
            if self.input_size is not None:
                if crop_size is not None:
                    self.center = [
                        me.y * self.input_size[0] / crop_size[0],
                        me.x * self.input_size[1] / crop_size[1],
                    ]
                else:
                    self.center = [me.y, me.x]
                if me.event == cv2.EVENT_MOUSEWHEEL:
                    if me.flags & cv2.EVENT_FLAG_CTRLKEY:
                        if me.flags > 0:
                            self.zoom *= 1.1
                        else:
                            self.zoom /= 1.1
                    else:
                        if me.flags > 0:
                            self.barrel_power *= 1.1
                        else:
                            self.barrel_power /= 1.1
                        print(self.barrel_power)

        self.mouse_control = m_loop
        return self

    def __call__(self, arr):
        """Run the lens distortion algorithm on the input."""
        zoom_out = 1.0 / self.zoom
        self._check_setup_bleed(arr)

        self.input_size = arr.shape

        y = np.arange(arr.shape[0])
        x = np.arange(arr.shape[1])
        if self._center is None:
            self._center = [len(y) / 2.0, len(x) / 2.0]
        y2_ = (y - (len(y) / 2.0)) * zoom_out / arr.shape[0]
        x2_ = (x - (len(x) / 2.0)) * zoom_out / arr.shape[1]
        p2 = np.array(np.meshgrid(x2_, y2_))

        cy = self._center[0] / arr.shape[0]
        cx = self._center[1] / arr.shape[1]

        theta = np.arctan2(p2[1], p2[0])

        radius = np.linalg.norm(p2, axis=0, ord=2)

        radius = pow(radius, self.barrel_power)

        x_new = 0.5 * (radius * np.cos(theta) + cx * 2)
        x_new = np.clip(x_new * len(x), 0, len(x) - 1)

        y_new = 0.5 * (radius * np.sin(theta) + cy * 2)
        y_new = np.clip(y_new * len(y), 0, len(y) - 1)

        p = np.array(np.meshgrid(y, x)).astype(np.uint32)

        p_new = np.array((y_new, x_new))
        p_new = p_new.astype(np.uint32)

        if self.use_bleed:
            arr2 = self.bleed.copy()
            self.run_bleed(arr2, x, y)
            arr2[p_new[0], p_new[1], :] = np.swapaxes(arr[p[0], p[1], :], 0, 1)
            self.bleed = arr2
        else:
            arr[p[0], p[1], :] = np.swapaxes(arr[p_new[0], p_new[1], :], 0, 1)

        return arr


class BarrelPyTorch(Barrel):
    """A barrel distortion callback class accelerated by PyTorch."""

    def __call__(self, arr):
        """Run a pytorch accelerated lens distortion algorithm on the input."""
        zoom_out = 1.0 / self.zoom
        self.input_size = arr.shape
        y = torch.arange(arr.shape[1]).type(torch.FloatTensor).cuda()
        x = torch.arange(arr.shape[0]).type(torch.FloatTensor).cuda()
        if self._center is None:
            self._center = [y.shape[0] / 2.0, x.shape[0] / 2.0]

        y2_ = (y - (y.shape[0] / 2.0)) * zoom_out / arr.shape[1]
        x2_ = (x - (x.shape[0] / 2.0)) * zoom_out / arr.shape[0]
        p2 = torch.stack(torch.meshgrid(x2_, y2_))

        cy = self._center[1] / arr.shape[1]
        cx = self._center[0] / arr.shape[0]

        theta = torch.atan2(p2[1], p2[0])

        radius = torch.norm(p2, dim=0)

        radius = torch.pow(radius, self.barrel_power)

        x_new = 0.5 * (radius * torch.cos(theta) + cx * 2)
        x_new = torch.clamp(x_new * x.shape[0], 0, x.shape[0] - 1)

        y_new = 0.5 * (radius * torch.sin(theta) + cy * 2)
        y_new = torch.clamp(y_new * y.shape[0], 0, y.shape[0] - 1)

        p = torch.stack(torch.meshgrid([x, y])).type(torch.IntTensor)

        p_new = torch.stack((x_new, y_new))
        p_new = p_new.type(torch.IntTensor)

        arr[p[0], p[1], :] = arr[p_new[0], p_new[1], :]

        return arr


class Mustache(ControllableLens):
    """A mustache distortion callback."""

    def __init__(
        self, use_bleed=False, barrel_power=1, pincushion_power=1, zoom=1, center=None
    ):
        """Create the mustache distortion callback."""
        super().__init__(use_bleed, zoom, center)
        self.center = center
        self.zoom = zoom
        self.use_bleed = use_bleed
        self.bleed = None
        self.barrel_power = barrel_power
        self.pincushion_power = pincushion_power
        self.mouse_control = None

    def enable_mouse_control(self):
        """
        Enable the default mouse loop to control the mustache distortion.

        ctrl+scroll wheel: zoom in and out
        shift+scroll wheel: adjust pincushion power
        scroll wheel: adjust barrel power
        """

        @mouse_loop
        def m_loop(me):
            self.center[:] = [me.y, me.x]
            if me.event == cv2.EVENT_MOUSEWHEEL:
                if me.flags & cv2.EVENT_FLAG_CTRLKEY:
                    if me.flags > 0:
                        self.zoom *= 1.1
                    else:
                        self.zoom /= 1.1
                elif me.flags & cv2.EVENT_FLAG_SHIFTKEY:
                    if me.flags > 0:
                        self.pincushion_power *= 1.1
                    else:
                        self.pincushion_power /= 1.1
                else:
                    if me.flags > 0:
                        self.barrel_power *= 1.1
                    else:
                        self.barrel_power /= 1.1

        self.mouse_control = m_loop

    def __call__(self, arr):
        """Run the mustache distortion algorithm on the input numpy array."""
        zoom_out = 1.0 / self.zoom
        self._check_setup_bleed(arr)

        y = np.arange(arr.shape[0])
        x = np.arange(arr.shape[1])
        if self.center is None:
            self.center = [len(y) / 2.0, len(x) / 2.0]
        y2_ = (y - self.center[0]) * zoom_out / arr.shape[0]
        x2_ = (x - self.center[1]) * zoom_out / arr.shape[1]
        p2 = np.array(np.meshgrid(x2_, y2_))

        barrel_power = self.barrel_power
        pincushion_power = self.pincushion_power

        theta = np.arctan2(p2[1], p2[0])

        radius = np.linalg.norm(p2, axis=0)
        radius2 = np.linalg.norm(p2, axis=0, ord=4)

        radius = pow(radius, barrel_power)
        radius2 = pow(radius2, pincushion_power)

        x_new = 0.5 * (radius2 * radius * np.cos(theta) + 1)
        x_new = np.clip(x_new * len(x), 0, len(x) - 1)

        y_new = 0.5 * (radius2 * radius * np.sin(theta) + 1)
        y_new = np.clip(y_new * len(y), 0, len(y) - 1)

        p = np.array(np.meshgrid(y, x)).astype(np.uint32)

        p_new = np.array((y_new, x_new)).astype(np.uint32)

        if self.use_bleed:
            arr2 = self.bleed.copy()
            self.run_bleed(arr2, x, y)
            arr2[p_new[0], p_new[1], :] = np.swapaxes(arr[p[0], p[1], :], 0, 1)
            self.bleed = arr2
        else:
            arr2 = np.zeros_like(arr)
            arr2[p_new[0], p_new[1], :] = np.swapaxes(arr[p[0], p[1], :], 0, 1)

        return arr2



"""Publish frames so any function within this program can find them."""

import asyncio
import sys
import threading
import time
import warnings

import cv2

using_pyv4l2cam = False
try:
    if sys.platform == "linux":
        from PyV4L2Cam.camera import Camera as pyv4lcamera  # type: ignore
        from PyV4L2Cam.controls import ControlIDs as pyv4lcontrolids  # type: ignore
        from PyV4L2Cam import convert_mjpeg, convert_rgb24  # type: ignore
        from PyV4L2Cam.get_camera import get_camera_by_bus_info, get_camera_by_string  # type: ignore

        using_pyv4l2cam = True
except ImportError:
    warnings.warn("Could not import PyV4L2Cam on linux. Camera capture will be slow.")
    warnings.warn(
        "To install, run: pip install git+https://github.com/simleek/PyV4L2Cam.git"
    )

import numpy as np

from displayarray.frame import subscriber_dictionary
from .np_to_opencv import NpCam
from displayarray._uid import uid_for_source

from typing import Union, Tuple, Optional, Dict, Any, List, Callable

FrameCallable = Callable[[np.ndarray], Optional[np.ndarray]]


def pub_cam_loop_pyv4l2(
    cam_id: Union[int, str, np.ndarray],
    request_size: Tuple[int, int] = (-1, -1),
    high_speed: bool = True,
    fps_limit: float = float("inf"),
):
    """
    Publish whichever camera you select to CVCams.<cam_id>.Vid, using v4l2 instead of opencv.

    You can send a quit command 'quit' to CVCams.<cam_id>.Cmd
    Status information, such as failure to open, will be posted to CVCams.<cam_id>.Status

    :param high_speed: Selects mjpeg transferring, which most cameras seem to support, so speed isn't limited
    :param fps_limit: Limits the frames per second.
    :param cam_id: An integer representing which webcam to use, or a string representing a video file.
    :param request_size: A tuple with width, then height, to request the video size.
    :return: True if loop ended normally, False if it failed somehow.
    """
    name = uid_for_source(cam_id)

    if isinstance(cam_id, (int, str)):
        if isinstance(cam_id, int):
            cam: pyv4lcamera = pyv4lcamera(  # type: ignore
                f"/dev/video{cam_id}", *request_size
            )
        else:
            if "usb" in cam_id:
                cam = get_camera_by_bus_info(cam_id, *request_size)  # type: ignore
            else:
                cam = get_camera_by_string(cam_id, *request_size)  # type: ignore
    else:
        raise TypeError(
            "Only strings or ints representing cameras are supported with v4l2."
        )

    subscriber_dictionary.register_cam(name, cam)

    sub = subscriber_dictionary.cam_cmd_sub(name)
    sub.return_on_no_data = ""
    msg = ""

    if high_speed and cam.pixel_format != "MJPEG":
        warnings.warn("Camera does not support high speed.")

    now = time.time()
    while msg != "quit":
        time.sleep(1.0 / (fps_limit - (time.time() - now)))
        now = time.time()
        frame_bytes = cam.get_frame()  # type: bytes

        if cam.pixel_format == "MJPEG":
            nd_frame = convert_mjpeg(frame_bytes)  # type: ignore
        elif cam.pixel_format == "RGB24":
            nd_frame = convert_rgb24(frame_bytes, cam.width, cam.height)  # type: ignore
        else:
            raise NotImplementedError(f"{cam.pixel_format} format not supported.")

        if nd_frame is not None:
            subscriber_dictionary.CV_CAMS_DICT[name].frame_pub.publish(nd_frame)
        else:
            cam.close()
            subscriber_dictionary.CV_CAMS_DICT[name].status_pub.publish("failed")
            return False

        msg = sub.get()
    sub.release()

    cam.close()
    return True


def pub_cam_loop_opencv(
    cam_id: Union[int, str, np.ndarray],
    request_size: Tuple[int, int] = (-1, -1),
    high_speed: bool = True,
    fps_limit: float = float("inf"),
    extra: Optional[List[Tuple[int, int]]] = None,
) -> bool:
    """
    Publish whichever camera you select to CVCams.<cam_id>.Vid.

    You can send a quit command 'quit' to CVCams.<cam_id>.Cmd
    Status information, such as failure to open, will be posted to CVCams.<cam_id>.Status

    :param high_speed: Selects mjpeg transferring, which most cameras seem to support, so speed isn't limited
    :param fps_limit: Limits the frames per second.
    :param cam_id: An integer representing which webcam to use, or a string representing a video file.
    :param request_size: A tuple with width, then height, to request the video size.
    :return: True if loop ended normally, False if it failed somehow.
    """
    name = uid_for_source(cam_id)

    if isinstance(cam_id, (int, str)):
        cam: Union[NpCam, cv2.VideoCapture] = cv2.VideoCapture(cam_id)
    elif isinstance(cam_id, np.ndarray):
        cam = NpCam(cam_id)
    else:
        raise TypeError(
            "Only strings or ints representing cameras, or numpy arrays representing pictures supported."
        )

    subscriber_dictionary.register_cam(name, cam)

    frame_counter = 0

    sub = subscriber_dictionary.cam_cmd_sub(name)
    sub.return_on_no_data = ""
    msg = ""

    if high_speed:
        try:
            cam.set(cv2.CAP_PROP_FOURCC, cv2.CAP_OPENCV_MJPEG)
        except AttributeError:
            warnings.warn("Please update OpenCV")

    cam.set(cv2.CAP_PROP_FRAME_WIDTH, request_size[0])
    cam.set(cv2.CAP_PROP_FRAME_HEIGHT, request_size[1])

    if not cam.isOpened():
        subscriber_dictionary.CV_CAMS_DICT[name].status_pub.publish("failed")
        return False
    now = time.time()
    while msg != "quit":
        time.sleep(1.0 / (fps_limit - (time.time() - now)))
        now = time.time()
        (ret, frame) = cam.read()  # type: Tuple[bool, np.ndarray ]
        if ret is False or not isinstance(frame, np.ndarray):
            cam.release()
            subscriber_dictionary.CV_CAMS_DICT[name].status_pub.publish("failed")
            return False
        if cam.get(cv2.CAP_PROP_FRAME_COUNT) > 0:
            frame_counter += 1
            if frame_counter >= cam.get(cv2.CAP_PROP_FRAME_COUNT):
                frame_counter = 0
                cam = cv2.VideoCapture(cam_id)
        subscriber_dictionary.CV_CAMS_DICT[name].frame_pub.publish(frame)
        msg = sub.get()
    sub.release()

    cam.release()
    return True


uid_dict: Dict[str, threading.Thread] = {}


def pub_cam_thread(
    cam_id: Union[int, str],
    request_ize: Tuple[int, int] = (-1, -1),
    high_speed: bool = True,
    fps_limit: float = float("inf"),
    force_backend="",
) -> threading.Thread:
    """Run pub_cam_loop in a new thread. Starts on creation."""

    name = uid_for_source(cam_id)
    if name in uid_dict.keys():
        t = uid_dict[name]
    else:
        if "cv" in force_backend.lower():
            pub_cam_loop = pub_cam_loop_opencv
        elif (
            sys.platform == "linux"
            and using_pyv4l2cam
            and (
                isinstance(cam_id, int)
                or (
                    isinstance(cam_id, str)
                    and any(["/dev/video" in cam_id, "usb" in cam_id])
                )
            )
        ) or "v4l2" in force_backend.lower():
            pub_cam_loop = pub_cam_loop_pyv4l2  # type: ignore
        else:
            pub_cam_loop = pub_cam_loop_opencv

        t = threading.Thread(
            target=pub_cam_loop, args=(cam_id, request_ize, high_speed, fps_limit)
        )
        uid_dict[name] = t
        t.start()
    return t


async def publish_updates_zero_mq(
    *vids,
    callbacks: Optional[
        Union[Dict[Any, FrameCallable], List[FrameCallable], FrameCallable]
    ] = None,
    fps_limit=float("inf"),
    size=(-1, -1),
    end_callback: Callable[[], bool] = lambda: False,
    blocking=False,
    publishing_address="tcp://127.0.0.1:5600",
    prepend_topic="",
    flags=0,
    copy=True,
    track=False,
):
    """Publish frames to ZeroMQ when they're updated."""
    import zmq
    from displayarray import read_updates

    ctx = zmq.Context()
    s = ctx.socket(zmq.PUB)
    s.bind(publishing_address)

    if not blocking:
        flags |= zmq.NOBLOCK

    try:
        for v in read_updates(vids, callbacks, fps_limit, size, end_callback, blocking):
            if v:
                for vid_name, frame in v.items():
                    md = dict(
                        dtype=str(frame.dtype),
                        shape=frame.shape,
                        name=prepend_topic + vid_name,
                    )
                    s.send_json(md, flags | zmq.SNDMORE)
                    s.send(frame, flags, copy=copy, track=track)
            if fps_limit:
                await asyncio.sleep(1.0 / fps_limit)
            else:
                await asyncio.sleep(0)
    except KeyboardInterrupt:
        pass
    finally:
        vid_names = [uid_for_source(name) for name in vids]
        for v in vid_names:
            subscriber_dictionary.stop_cam(v)


async def publish_updates_ros(
    *vids,
    callbacks: Optional[
        Union[Dict[Any, FrameCallable], List[FrameCallable], FrameCallable]
    ] = None,
    fps_limit=float("inf"),
    size=(-1, -1),
    end_callback: Callable[[], bool] = lambda: False,
    blocking=False,
    node_name="displayarray",
    publisher_name="npy",
    rate_hz=None,
    dtype=None,
):
    """Publish frames to ROS when they're updated."""
    import rospy
    from rospy.numpy_msg import numpy_msg
    import std_msgs.msg
    from displayarray import read_updates

    def get_msg_type(dtype):
        if dtype is None:
            msg_type = {
                np.float32: std_msgs.msg.Float32(),
                np.float64: std_msgs.msg.Float64(),
                np.bool: std_msgs.msg.Bool(),
                np.char: std_msgs.msg.Char(),
                np.int16: std_msgs.msg.Int16(),
                np.int32: std_msgs.msg.Int32(),
                np.int64: std_msgs.msg.Int64(),
                np.str: std_msgs.msg.String(),
                np.uint16: std_msgs.msg.UInt16(),
                np.uint32: std_msgs.msg.UInt32(),
                np.uint64: std_msgs.msg.UInt64(),
                np.uint8: std_msgs.msg.UInt8(),
            }[dtype]
        else:
            msg_type = (
                dtype  # allow users to use their own custom messages in numpy arrays
            )
        return msg_type

    publishers: Dict[str, rospy.Publisher] = {}
    rospy.init_node(node_name, anonymous=True)
    try:
        for v in read_updates(vids, callbacks, fps_limit, size, end_callback, blocking):
            if v:
                if rospy.is_shutdown():
                    break
                for vid_name, frame in v.items():
                    if vid_name not in publishers:
                        dty = frame.dtype if dtype is None else dtype
                        publishers[vid_name] = rospy.Publisher(
                            publisher_name + vid_name,
                            numpy_msg(get_msg_type(dty)),
                            queue_size=10,
                        )
                    publishers[vid_name].publish(frame)
            if rate_hz:
                await asyncio.sleep(1.0 / rate_hz)
            else:
                await asyncio.sleep(0)
    except KeyboardInterrupt:
        pass
    finally:
        vid_names = [uid_for_source(name) for name in vids]
        for v in vid_names:
            subscriber_dictionary.stop_cam(v)
    if rospy.core.is_shutdown():
        raise rospy.exceptions.ROSInterruptException("rospy shutdown")
